#    Quizpy - Creating Moodle exams in Python
#    Copyright (C) 2021  Sebastian Bräuer
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.

import xml.etree.ElementTree as ET
import typing

from dataclasses import dataclass, field

from .common import Question
from .multipe_choice import Choice


@dataclass
class ShortAnswer(Question):
    case_sensitive: bool = False
    accepted_answers: typing.List[Choice] = field(default_factory=list)

    def to_xml(self) -> ET.Element:
        node = self.generate_common_xml('shortanswer')
        ET.SubElement(node, 'usecase').text = '1' if self.case_sensitive else 0

        for answer in self.accepted_answers:
            node.append(answer.to_xml())

        return node

    def to_cloze(self, weight: int) -> str:
        cloze = f"{weight}:SA:" if not self.case_sensitive else f"{weight}:SAC:"

        cloze += '~'.join(map(lambda s: s.to_cloze(), self.accepted_answers))

        return f"{{{cloze}}}"


class Description(Question):

    def to_xml(self) -> ET.Element:
        node = self.generate_common_xml('description')
        return node


@dataclass
class Essay(Question):
    grading_info: str = ""
    response_format: str = "editor"
    response_required: bool = True
    response_template: str = ""
    attachements_required: bool = False
    max_attachements: int = 0
    max_lines: int = 15

    def to_xml(self) -> ET.Element:
        node = self.generate_common_xml('essay')

        ET.SubElement(node, 'responseformat').text = self.response_format
        ET.SubElement(node, 'responserequired').text = '1' if self.response_required else '0'
        ET.SubElement(node, 'responsefieldlines').text = str(self.max_lines)
        ET.SubElement(node, 'attachements').text = str(self.max_attachements)
        ET.SubElement(node, 'attachementsrequired').text = '1' if self.attachements_required else '0'

        grading_info = ET.SubElement(node, 'graderinfo')
        grading_info.set('format', self.format)
        ET.SubElement(grading_info, 'text').text = self.grading_info

        template = ET.SubElement(node, 'responsetemplate')
        template.set('format', self.format)
        ET.SubElement(template, 'text').text = self.response_template

        return node


class Cloze(Question):
    """The question text will be rendered using the custom Cloze syntax.
    See https://docs.moodle.org/310/en/Embedded_Answers_(Cloze)_question_type.

    You can also use the `to_cloze()` functions of the relevant question classes to generate the syntax for you.
    """

    def to_xml(self) -> ET.Element:
        node = self.generate_common_xml('cloze')
        return node



#    Quizpy - Creating Moodle exams in Python
#    Copyright (C) 2021  Sebastian Bräuer
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.

import xml.etree.ElementTree as ET
import typing
import enum

from dataclasses import dataclass, field

from .common import Question
from .multipe_choice import Choice


class UnitGradingType(enum.Enum):
    NO_PENALTY = 0
    FRAC_RESPONSE_GRADE = 1
    FRAC_TOTAL_GRADE = 2


class UnitBehaviour(enum.Enum):
    Optional = 0
    Force = 1
    NoUnits = 3


@dataclass
class NumericalAnswer(Choice):
    tolerance: float = 0.0

    def to_xml(self) -> ET.Element:
        answer = super().to_xml()
        ET.SubElement(answer, 'tolerance').text = str(self.tolerance)
        return answer

    def to_cloze(self) -> str:
        if self.fraction == 100.0:
            s = f"={self.text}:{self.tolerance}#{self.feedback}"
        elif self.fraction == 0.0:
            s = f"{self.text}:{self.tolerance}#{self.feedback}"
        else:
            s = f"%{int(self.fraction)}%{self.text}:{self.tolerance}#{self.feedback}"
        return s


@dataclass
class Numerical(Question):
    accepted_answers: typing.List[NumericalAnswer] = field(default_factory=list)
    units: typing.List[typing.Tuple[str, float]] = field(default_factory=list)
    units_left: bool = False
    unit_grading_type: UnitGradingType = UnitGradingType.NO_PENALTY
    unit_behaviour: UnitBehaviour = UnitBehaviour.NoUnits
    unit_penalty: float = 0.1

    # TODO: Eingabeformat wird im XML komplett vergessen

    def to_xml(self) -> ET.Element:
        node = self.generate_common_xml('numerical')

        for answer in self.accepted_answers:
            node.append(answer.to_xml())

        units = ET.SubElement(node, 'units')
        for unit, multiplier in self.units:
            u = ET.SubElement(units, 'unit')
            ET.SubElement(u, 'multiplier').text = str(multiplier)
            ET.SubElement(u, 'unit_name').text = unit

        ET.SubElement(node, 'unitgradingtype').text = str(self.unit_grading_type.value)
        ET.SubElement(node, 'unitsleft').text = '1' if self.units_left else '0'
        ET.SubElement(node, 'showunits').text = str(self.unit_behaviour.value)
        ET.SubElement(node, 'unitpenalty').text = str(self.unit_penalty)

        return node

    def to_cloze(self, weight: int):
        cloze = f'{weight}:NM:'
        cloze += '~'.join(map(lambda a: a.to_cloze(), self.accepted_answers))

        return f'{{{cloze}}}'




#    Quizpy - Creating Moodle exams in Python
#    Copyright (C) 2021  Sebastian Bräuer
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.

import xml.etree.ElementTree as ET
import typing

from dataclasses import dataclass, field

from .common import Question


@dataclass
class Choice:
    text: str
    fraction: float

    feedback: str = ''
    format: str = 'html'

    def to_xml(self) -> ET.Element:
        answer = ET.Element('answer')
        answer.set('fraction', str(self.fraction))
        answer.set('format', self.format)

        ET.SubElement(answer, 'text').text = self.text
        fb = ET.SubElement(answer, 'feedback')
        fb.set('format', self.format)
        ET.SubElement(fb, 'text').text = self.feedback

        return answer

    def to_cloze(self) -> str:
        if self.fraction == 100.0:
            s = f"={self.text}#{self.feedback}"
        elif self.fraction == 0.0:
            s = f"{self.text}#{self.feedback}"
        else:
            s = f"%{int(self.fraction)}%{self.text}#{self.feedback}"
        return s


@dataclass
class MultipleChoice(Question):
    shuffle: bool = True
    single: bool = False
    numbering: str = 'abc'
    choices: typing.List[Choice] = field(default_factory=list)

    correct_feedback: str = 'Die Antwort ist richtig.'
    partial_feedback: str = 'Die Antwort ist teilweise richtig.'
    incorrect_feedback: str = 'Die Antwort ist falsch.'

    def add_choice(self, text: str, fraction: float, feedback: str = '', format: str = 'html'):
        c = Choice(text, fraction,feedback,format)
        self.choices.append(c)
        return c

    def to_xml(self):
        node = self.generate_common_xml('multichoice')
        node.set('type', 'multichoice')

        ET.SubElement(node, 'shuffleanswers').text = 'true' if self.shuffle else 'false'

        ET.SubElement(node, 'single').text = 'true' if self.single else 'false'
        ET.SubElement(node, 'answernumbering').text = self.numbering

        correct_feedback = ET.SubElement(node, 'correctfeedback')
        correct_feedback.set('format', self.format)
        ET.SubElement(correct_feedback, 'text').text = self.correct_feedback

        partial_feedback = ET.SubElement(node, 'partiallycorrectfeedback')
        partial_feedback.set('format', self.format)
        ET.SubElement(partial_feedback, 'text').text = self.partial_feedback

        wrong_feedback = ET.SubElement(node, 'incorrectfeedback')
        wrong_feedback.set('format', self.format)
        ET.SubElement(wrong_feedback, 'text').text = self.incorrect_feedback

        for c in self.choices:
            node.append(c.to_xml())

        return node

    def to_cloze(self, weight: int, display_type: str = "dropdown", alignment: str = 'horizontal', shuffle: bool = False) -> str:

        if display_type not in {'dropdown', 'radio', 'checkbox'}:
            raise ValueError("display_type must be either 'dropdown', 'radio' or 'checkbox'!")

        if alignment not in {'horizontal', 'vertical'}:
            raise ValueError("alignment must be either 'horizontal', 'vertical'!")

        if display_type == 'dropdown':
            cloze_type = 'MC'
        elif display_type == 'radio':
            if alignment == 'horizontal':
                cloze_type = 'MCH'
            else:
                cloze_type = 'MCV'
        else:
            if alignment == 'horizontal':
                cloze_type = 'MRH'
            else:
                cloze_type = 'MR'

        if shuffle:
            cloze_type += 'S'

        cloze = f'{weight}:{cloze_type}:'
        cloze += '~'.join(map(lambda s: s.to_cloze(), self.choices))

        return f"{{{cloze}}}"


@dataclass
class MultiTrueFalse(Question):
    scoring_method: str = 'subpoints'
    shuffle: bool = True
    options: typing.List = field(default_factory=lambda: ['Wahr', 'Falsch'])
    statements: typing.List[typing.Tuple[str, str, str]] = field(default_factory=list)

    def add_statement(self, text: str, value: str, feedback: str = ''):
        self.statements.append((text, value, feedback))

    def to_xml(self) -> ET.Element:
        node = self.generate_common_xml('mtf')

        scoring = ET.SubElement(node, 'scoringmethod')
        ET.SubElement(scoring, 'text').text = self.scoring_method

        ET.SubElement(node, 'shuffleanswers').text = 'true' if self.shuffle else 'false'

        num_rows = len(self.statements)
        num_cols = len(self.options)

        ET.SubElement(node, 'numberofrows').text = str(num_rows)
        ET.SubElement(node, 'numberofcolumns').text = str(num_cols)

        for i, (text, _, feedback) in enumerate(self.statements, start=1):
            row = ET.SubElement(node, 'row')
            row.set('number', str(i))

            opt = ET.SubElement(row, 'optiontext')
            opt.set('format', self.format)
            ET.SubElement(opt, 'text').text = text

            fb = ET.SubElement(row, 'feedbacktext')
            fb.set('format', self.format)
            ET.SubElement(fb, 'text').text = feedback

        for i, opt in enumerate(self.options, start=1):
            col = ET.SubElement(node, 'column')
            col.set('number', str(i))

            resp = ET.SubElement(col, 'responsetext')
            resp.set('format', self.format)
            ET.SubElement(resp, 'text').text = opt

        for i, (_, correct, _) in enumerate(self.statements, start=1):
            for j, opt in enumerate(self.options, start=1):
                weight = ET.SubElement(node, 'weight')
                weight.set('rownumber', str(i))
                weight.set('columnnumber', str(j))

                ET.SubElement(weight, 'value').text = '1.000' if opt == correct else '0.000'

        return node


@dataclass
class Matching(Question):
    shuffle: bool = True
    pairs: typing.List[typing.Tuple[str, str]] = field(default_factory=list)

    def to_xml(self) -> ET.Element:
        node = self.generate_common_xml('matching')

        for text, answer in self.pairs:
            subquestion = ET.SubElement(node, 'subquestion')
            subquestion.set('format', self.format)

            ET.SubElement(subquestion, 'text').text = text
            aw = ET.SubElement(subquestion, 'answer')
            ET.SubElement(aw, 'text').text = answer

        return node




#    Quizpy - Creating Moodle exams in Python
#    Copyright (C) 2021  Sebastian Bräuer
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.

import xml.etree.ElementTree as ET
import typing

from dataclasses import dataclass, field
from .common import Question, InlineFile

# TODO Question Type ddwtos

@dataclass
class DragOption:
    text: str
    group: int = 1
    infinite: bool = False


@dataclass
class DropZone:
    location: typing.Tuple[int, int]
    choice: DragOption
    text: str = ""


@dataclass
class DDImage(Question):

    base_image: InlineFile = None  # Unfortunately,  this must be set with a default
    options: typing.List[DragOption] = field(default_factory=list)
    drops: typing.List[DropZone] = field(default_factory=list)
    shuffle: bool = False

    def to_xml(self) -> ET.Element:

        if self.base_image is None:
            raise ValueError("You need to set a base image!")

        node = self.generate_common_xml('ddimageortext')

        if self.shuffle:
            ET.SubElement(node, "shuffleanswers")

        node.append(self.base_image.to_xml())

        for i, opt in enumerate(self.options, start=1):
            drag = ET.SubElement(node, 'drag')
            ET.SubElement(drag, 'text').text = opt.text
            ET.SubElement(drag, 'draggroup').text = str(opt.group)
            ET.SubElement(drag, 'no').text = str(i)

            if opt.infinite:
                ET.SubElement(drag, 'infinite')

        for i, d in enumerate(self.drops, start=1):
            x, y = d.location
            drop = ET.SubElement(node, 'drop')
            ET.SubElement(drop, 'text')  # Empty sub-element
            ET.SubElement(drop, 'xleft').text = str(x)
            ET.SubElement(drop, 'ytop').text = str(y)
            ET.SubElement(drop, 'no').text = str(i)

            try:
                ET.SubElement(drop, 'choice').text = str(self.options.index(d.choice) + 1)
            except ValueError:
                raise ValueError("Choice for dropzone not found in drag options!")

        return node



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="quizpy",
    version="0.0.2",
    author="Sebastian Bräuer",
    author_email="braeuer@tu-berlin.de",
    description="A package to create Moodle exams using Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://git.tu-berlin.de/tkn/quizpy",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.6",
)
